package port

import (
	"net/http"
	"strconv"

	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"gitea.dwysokinski.me/twhelp/corev3/internal/port/internal/apimodel"
)

//nolint:gocyclo
func (h *apiHTTPHandler) ListPlayers(
	w http.ResponseWriter,
	r *http.Request,
	_ apimodel.VersionCodePathParam,
	serverKey apimodel.ServerKeyPathParam,
	params apimodel.ListPlayersParams,
) {
	domainParams := domain.NewListPlayersParams()

	if err := domainParams.SetSort([]domain.PlayerSort{domain.PlayerSortIDASC}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListPlayersErrorPath).render(w, r, err)
		return
	}

	if params.Sort != nil {
		if err := domainParams.PrependSortString(*params.Sort); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayersErrorPath).render(w, r, err)
			return
		}
	}

	if err := domainParams.SetServerKeys([]string{serverKey}); err != nil {
		h.errorRenderer.withErrorPathFormatter(formatListPlayersErrorPath).render(w, r, err)
		return
	}

	if params.Deleted != nil {
		if err := domainParams.SetDeleted(domain.NullBool{
			Value: *params.Deleted,
			Valid: true,
		}); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayersErrorPath).render(w, r, err)
			return
		}
	}

	if params.Limit != nil {
		if err := domainParams.SetLimit(*params.Limit); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayersErrorPath).render(w, r, err)
			return
		}
	}

	if params.Cursor != nil {
		if err := domainParams.SetEncodedCursor(*params.Cursor); err != nil {
			h.errorRenderer.withErrorPathFormatter(formatListPlayersErrorPath).render(w, r, err)
			return
		}
	}

	res, err := h.playerSvc.List(r.Context(), domainParams)
	if err != nil {
		h.errorRenderer.render(w, r, err)
		return
	}

	renderJSON(w, r, http.StatusOK, apimodel.NewListPlayersResponse(res))
}

func formatListPlayersErrorPath(segments []errorPathSegment) []string {
	if segments[0].model != "ListPlayersParams" {
		return nil
	}

	switch segments[0].field {
	case "cursor":
		return []string{"$query", "cursor"}
	case "limit":
		return []string{"$query", "limit"}
	case "deleted":
		return []string{"$query", "deleted"}
	case "sort":
		path := []string{"$query", "sort"}
		if segments[0].index >= 0 {
			path = append(path, strconv.Itoa(segments[0].index))
		}
		return path
	default:
		return nil
	}
}
